@@ -1,46 +1,48 @@
 #!/usr/bin/env python3
-# This package contains functions for drawing basic geometric shapes, like
-# lines and triangles, into 3D numpy arrays (and the code almost works for nD
-# arrays). opencv has drawline and drawtriangle functions, but those only
-# operate on 2D image arrays. scipy.ndimage provides many other helpful,
-# related functions for working with nD image arrays.
+"""
+This package contains functions for drawing basic geometric shapes, like
+lines and triangles, into 3D numpy arrays (and the code almost works for nD
+arrays). opencv has drawline and drawtriangle functions, but those only
+operate on 2D image arrays. scipy.ndimage provides many other helpful,
+related functions for working with nD image arrays.
 
-# Subpixel indexing conventions:
-#
-# 'corner' convention (default): An operation that tries to get or set a pixel
-# located at (x, y, z) will actually get or set the pixel with index (floor(x),
-# floor(y), floor(z)). Therefore the pixel at location (x, y, z) represents the
-# cube of physical locations (a, b, c) such that a is in [x, x+1), b is in [y,
-# y+1), and c is [z, z+1). This is called the 'corner' convention because
-# integer locations like (5, 2, 10) point to the corner of a voxel.
-#
-# 'center' convention: An operation that tries to get or set a pixel located at
-# (x, y, z) will actually get or set the pixel with index (round(x), round(y),
-# round(z)). Therefore the pixel at location (x, y, z) represents the cube of
-# physical locations (a, b, c) such that a is in [x-0.5, x+0.5), b is in
-# [y-0.5, y+0.5), and c is in [z-0.5, z+0.5). This is called the 'center'
-# convention because integer locations like (5, 2, 10) point to the middle of a
-# voxel. This convention can be selected by passing convention='center' to any
-# of the functions in this package (once I finish implementing it).
-# 
-# Functions in this package use non-integer data types to preserve accuracy,
-# only flooring or rounding (depending on the convention) during get-pixel or
-# set-pixel operations.
+Subpixel indexing conventions:
+
+'corner' convention (default): An operation that tries to get or set a pixel
+located at (x, y, z) will actually get or set the pixel with index (floor(x),
+floor(y), floor(z)). Therefore the pixel at location (x, y, z) represents the
+cube of physical locations (a, b, c) such that a is in [x, x+1), b is in [y,
+y+1), and c is [z, z+1). This is called the 'corner' convention because
+integer locations like (5, 2, 10) point to the corner of a voxel.
+
+'center' convention: An operation that tries to get or set a pixel located at
+(x, y, z) will actually get or set the pixel with index (round(x), round(y),
+round(z)). Therefore the pixel at location (x, y, z) represents the cube of
+physical locations (a, b, c) such that a is in [x-0.5, x+0.5), b is in
+[y-0.5, y+0.5), and c is in [z-0.5, z+0.5). This is called the 'center'
+convention because integer locations like (5, 2, 10) point to the middle of a
+voxel. This convention can be selected by passing convention='center' to any
+of the functions in this package (once I finish implementing it).
+
+Functions in this package use non-integer data types to preserve accuracy,
+only flooring or rounding (depending on the convention) during get-pixel or
+set-pixel operations.
 
 
-# LIST OF MAJOR TODOS
+LIST OF MAJOR TODOS
 
-# Implement 'center' convention logic for all functions.
+Implement 'center' convention logic for all functions.
 
-# Make the image argument optional for drawpoint, drawline, and drawtriangle,
-# in which case they should create an image just large enough to hold the drawn
-# object and then return that image
+Make the image argument optional for drawpoint, drawline, and drawtriangle,
+in which case they should create an image just large enough to hold the drawn
+object and then return that image
 
-# Make get_voxels_within_distance be able to take a multidimensional distance
-# argument, in which case it should return all the voxels within an ellipse
-# having its radii along each axis specified by the elements of distance. This
-# is an extension of the current functionality, since currently a sphere is
-# returned with radius in all dimensions equal to the int/float distance.
+Make get_voxels_within_distance be able to take a multidimensional distance
+argument, in which case it should return all the voxels within an ellipse
+having its radii along each axis specified by the elements of distance. This
+is an extension of the current functionality, since currently a sphere is
+returned with radius in all dimensions equal to the int/float distance.
+"""
 
 
 import math
@@ -49,7 +51,7 @@ import itertools
 import numpy as np
 np.set_printoptions(suppress=True)
 
-from .utils import *
+from .utils import eq, ifloor, iceil, iround, is_out_of_bounds, remove_out_of_bounds
 
 
 # --- Primitive shapes - points, lines, triangles, circles, spheres --- #
@@ -72,7 +74,7 @@ def drawline(image, pt1, pt2, value, thickness=1,
     https://www.tutorialspoint.com/computer_graphics/line_generation_algorithm.htm
     with some additional logic that allows for non-integer point coordinates.
     """
-    pt1, pt2, long_axis = preprocess_polygon_vertices(pt1, pt2)
+    pt1, pt2, long_axis = _preprocess_polygon_vertices(pt1, pt2)
     #print(f'after preprocessing, pt1 is {pt1}, pt2 is {pt2}, and long_axis is {long_axis}')
 
     # Deal with the start and end points as special - always draw them
@@ -123,12 +125,13 @@ def drawline(image, pt1, pt2, value, thickness=1,
     # slightly (20%) faster than calling drawpoint - despite drawpoint
     # literally just calling thicken and then imset itself. ???
     pts = thicken(pts, thickness)
-    imset(image, pts, value, 
+    imset(image, pts, value,
           convention=convention, out_of_bounds=out_of_bounds)
 
 
 # TODO switch to kwargs
 def drawtriangle(image, pt1, pt2, pt3, value, thickness=1,
+                 watertight=True,
                  fill_value=None, convention='corner', out_of_bounds='ignore'):
 
     if fill_value is None:
@@ -140,7 +143,7 @@ def drawtriangle(image, pt1, pt2, pt3, value, thickness=1,
                  convention=convention, out_of_bounds=out_of_bounds)
         return
 
-    pt1, pt2, pt3, long_axis = preprocess_polygon_vertices(pt1, pt2, pt3)
+    pt1, pt2, pt3, long_axis = _preprocess_polygon_vertices(pt1, pt2, pt3)
 
     vec_1to3 = pt3 - pt1
     if eq(vec_1to3[long_axis], 0):
@@ -179,7 +182,7 @@ def drawtriangle(image, pt1, pt2, pt3, value, thickness=1,
 
         # To ensure 3D triangles end up water-tight, need the thickness of the
         # filler lines must be >1 in at least one dimension.
-        if all(thickness == 1):
+        if watertight and all(thickness == 1):
             slopes = abs(endpt - basept)
             if all(eq(slopes, 0)):
                 pass
@@ -216,7 +219,7 @@ def drawcircle(image, center, perpendicular, radius, value,
     Draw a circle
     See drawpoint for all kwargs.
     """
-    ndims = 3  #TODO code up logic for determining this from the input
+    ndims = 3  # TODO code up logic for determining this from the input
     return_image = False
     if image is None and center is None:
         image = np.zeros((2*iceil(radius) + 1,)*ndims, dtype=np.uint8)
@@ -355,40 +358,28 @@ def imget(image, coords, convention='corner',
     makes more sense in most graphics applications.
         out_of_bounds, string   : 'ignore' (default), 'wrap', or 'raise'
     """
-    assert convention in ['corner', 'center']
     if not isinstance(coords, np.ndarray):
         coords = np.array(coords)
-    ax = len(coords.shape) - 1
+    n_original = coords.shape[0]
 
-    if convention == 'corner':
-        is_negative = coords < 0
-        exceeds_dimensions = coords >= image.shape
-    elif convention == 'center':
-        is_negative = coords < -0.5
-        exceeds_dimensions = coords + 0.5 >= image.shape
-
-    is_out_of_bounds = np.logical_or(is_negative, exceeds_dimensions)
-    if (is_out_of_bounds).any():
-        def warn():
+    if out_of_bounds == 'wrap':
+        allow_negative_wrapping = True
+    elif out_of_bounds == 'ignore':
+        allow_negative_wrapping = False
+    elif out_of_bounds != 'raise':
+        raise ValueError("out_of_bounds must be 'ignore', 'raise', or"
+                         f"'wrap' but was {out_of_bounds}.")
+    coords = remove_out_of_bounds(coords, image.shape, allow_negative_wrapping, convention)
+    if n_original > coords.shape[0]:
+        n_out_of_bounds = n_original - coords.shape[0]
+        if out_of_bounds == 'raise':
+            raise IndexError(f'{n_out_of_bounds}/{n_original} coordinates'
+                             'are out of bounds.')
+        if verbose:
             print('WARNING: Some requested coordinates are out of bounds.'
                   ' The returned list of values will be shorter than the'
                   ' request list, and therefore the returned values will not'
                   ' match 1-to-1 with the requested coordinates.')
-        if out_of_bounds == 'ignore':
-            coords = coords[~is_out_of_bounds.any(axis=ax)]
-            if verbose:
-                print(f'verbose={verbose}')
-                warn()
-        elif out_of_bounds == 'wrap':
-            coords = coords[~exceeds_dimensions.any(axis=ax)]
-            if verbose:
-                warn()
-        elif out_of_bounds == 'raise':
-            raise IndexError(f'Some coordinates out of bounds:\n'
-                             '{coords[is_out_of_bounds]}')
-        else:
-            raise ValueError("out_of_bounds must be 'ignore', 'raise', or"
-                             f"'wrap' but was {out_of_bounds}.")
 
     if convention == 'corner':
         return image[tuple(ifloor(coords).T)]
@@ -406,26 +397,20 @@ def imset(image, coords, value, convention='corner', out_of_bounds='ignore'):
     makes more sense in most graphics applications.
         out_of_bounds, string   : 'ignore' (default), 'wrap', or 'raise'
     """
-    assert convention in ['corner', 'center']
     if not isinstance(coords, np.ndarray):
         coords = np.array(coords)
 
-    if convention == 'corner':
-        is_negative = coords < 0
-        exceeds_dimensions = coords >= image.shape
-    else:  # convention == 'center'
-        is_negative = coords < -0.5
-        exceeds_dimensions = coords + 0.5 >= image.shape
-
-    is_out_of_bounds = np.logical_or(is_negative, exceeds_dimensions)
-    if (is_out_of_bounds).any():
-        if out_of_bounds == 'ignore':
-            coords = coords[~is_out_of_bounds.any(axis=1)]
-        elif out_of_bounds == 'wrap':
-            coords = coords[~exceeds_dimensions.any(axis=1)]
+    if out_of_bounds == 'wrap':
+        allow_negative_wrapping = True
+    else:
+        allow_negative_wrapping = False
+    is_oob = is_out_of_bounds(coords, image.shape, allow_negative_wrapping, convention)
+    if is_oob.any():
+        if out_of_bounds in ['ignore', 'wrap']:
+            coords = coords[~is_oob]
         elif out_of_bounds == 'raise':
-            raise IndexError(f'Some coordinates out of bounds:\n'
-                             '{coords[is_out_of_bounds]}')
+            raise IndexError('Some coordinates out of bounds:\n'
+                             f'{coords[is_oob]}')
         else:
             raise ValueError("out_of_bounds must be 'ignore', 'raise', or"
                              f"'wrap' but was {out_of_bounds}.")
@@ -478,24 +463,24 @@ def drawmesh(image, mesh_fn, value=255, verbose=True,
     if verbose:
         for i in range(n):
             print(f'Drawing triangle {i+1} of {n}')
-            drawtriangle(image, v0[i], v1[i], v2[i], value, fill_value=value,
-                    **kwargs)
+            drawtriangle(image, v0[i], v1[i], v2[i], value,
+                         fill_value=value, **kwargs)
     else:
         for i in range(n):
-            drawtriangle(image, v0[i], v1[i], v2[i], value, fill_value=value,
-                    **kwargs)
+            drawtriangle(image, v0[i], v1[i], v2[i], value,
+                         fill_value=value, **kwargs)
 
 
 # --- Misc graphics operations --- #
 suffixes = {1: 'st', 2: 'nd', 3: 'rd'}
-def preprocess_polygon_vertices(*pts):
+def _preprocess_polygon_vertices(*pts):
     """
     Given a list of polygon vertices, find the axis along which the polygon is
     longest, and order the points by their position along that axis.
     returns a tuple containing the ordered points plus an integer indicating
     the longest axis.
     Usage example:
-        pt1, pt2, long_axis = preprocess_polygon_vertices(image, pt1, pt2)
+        pt1, pt2, long_axis = _preprocess_polygon_vertices(image, pt1, pt2)
     """
     # TODO TODO TODO make this aware of voxel sizes!! If voxel size is
     # anisotropic, the longest axis according to its physical space measurement
@@ -551,7 +536,10 @@ def thicken(pts, thickness=1, no_duplicates=True):
     if len(pts.shape) == 1:
         pts = np.expand_dims(pts, axis=0)
 
-    if thickness is 1:
+    if isinstance(thickness, np.ndarray):
+        if (thickness == 1).all():
+            return pts
+    elif thickness in (1, (1, 1, 1), [1, 1, 1]):
         return pts
 
     try:
@@ -567,13 +555,13 @@ def thicken(pts, thickness=1, no_duplicates=True):
         #offsets = get_voxels_within_distance(thickness)
     #else:
     offsets = np.array(list(itertools.product(*[range(-i//2 + 1, i//2 + 1)
-                                       for i in thickness])))
+                                                for i in thickness])))
 
     new_pts = np.concatenate([pts + offset for offset in offsets])
 
-
-    if no_duplicates: #TODO test if the added time here is worth the savings
-                      #when imsetting the resulting points
+    # TODO test if the added time here is worth the time savings
+    # when imsetting the resulting points
+    if no_duplicates:
         new_pts = np.unique(new_pts, axis=0)
 
     return new_pts
@@ -595,9 +583,9 @@ def get_voxels_within_distance(distance, center=None, ndims=None,
     assert metric in ['euclidian', 'manhattan', 'chebyshev']
 
     if ndims is None:
-        if center != None:
+        if center is not None:
             ndims = len(center)
-        elif voxel_size != None:
+        elif voxel_size is not None:
             ndims = len(voxel_size)
         else:
             ndims = 3
@@ -636,8 +624,8 @@ def get_voxels_within_distance(distance, center=None, ndims=None,
     }
     d = norms[metric]
 
-    candidate_pts_scaled = candidate_pts * voxel_size
-    return candidate_pts[d(candidate_pts_scaled) <= distance] + center
+    candidate_pts_scaled = candidate_pts * voxel_size / distance
+    return candidate_pts[d(candidate_pts_scaled) <= 1] + center
 
 
 def floodfill(image, seed, fill_value, fill_diagonally=False,
@@ -645,15 +633,14 @@ def floodfill(image, seed, fill_value, fill_diagonally=False,
     """
     This implementation is very slow and needs improvement
     """
-    import npimage
     seed_value = image[seed]
     if verbose: print(f'seed_value={seed_value}')
     if fill_diagonally:
         neighbors = get_voxels_within_distance(1, ndims=len(image.shape),
-                metric='chebyshev')
+                                               metric='chebyshev')
     else:
         neighbors = get_voxels_within_distance(1, ndims=len(image.shape))
-        
+
     if verbose: print(f'neighbors={neighbors}')
     neighbors = neighbors[~(neighbors == 0).all(axis=1)]
     fill_mask = np.zeros(image.shape, dtype=bool)
@@ -695,12 +682,12 @@ def floodfill(image, seed, fill_value, fill_diagonally=False,
             and (neighbor < image.shape).all()
             and eq(imget(image, neighbor), seed_value, tolerance=tolerance)
             and not imget(fill_mask, neighbor)]
-        for coord in wavefront]
+         for coord in wavefront]
 
         if debug_frequency is not None and iteration % debug_frequency == 0:
             print(f'Writing iteration{iteration}.nrrd')
-            npimage.numpy_to_imagefile(image, f'iteration{iteration}.nrrd')
-
+            from .core import save
+            save(image, f'iteration{iteration}.nrrd')
 
         #recurseifneighbors
         if len(new_wavefront) > 0:
@@ -719,7 +706,7 @@ def to_physical_coordinates(voxel_coordinates, voxel_size, offset=0):
     (in units of microns per voxel) and offset (indicating the physical
     coordinates of the origin voxel, in microns)
     """
-    #TODO test this. Haven't yet run it.
+    #TODO test this.
     return voxel_coordinates * voxel_size + offset
 
 
@@ -736,6 +723,3 @@ def to_voxel_coordinates(physical_coordinates, voxel_size, offset=0):
     #print(f'offset={offset}')
     #print(f'voxel_size={voxel_size}')
     return (physical_coordinates - offset) / voxel_size
-
-
-
